MODULE Controls;

IMPORT
  SYSTEM, W := Windows, Cl := Classes;

CONST
  CD_WINDOW*   = {1};
  CD_CONTROL*  = {2};

  sfcaption  = 1;
  sfenabled  = 2;
  sfvisible  = 3;
  sfsize     = 4;
  sfparent   = 5;
  sfstyle    = 6;
  sfexstyle  = 7;
  sfcolor    = 8;

  ssShift    = 1;
  ssAlt	     = 2;
  ssCtrl     = 3;
  ssLeft     = 4;
  ssRight    = 5;
  ssMiddle   = 6;
  ssDouble   = 7;

  mbLeft     = 1;
  mbRight    = 2;
  mbMiddle   = 3;

  alTop*     = 1;
  alLeft*    = 2;
  alRight*   = 3;
  alBottom*  = 4;
  alClient*  = 5;

VAR
  ControlList*: TList;
  MainForm*: TControl;

TYPE
  Char         = W.Char;
  Boolean      = W.Boolean;
  ShortInt     = W.ShortInt;
  Integer      = W.Integer;
  Longint      = W.Longint;
  Shortreal    = W.Shortreal;
  Real         = W.Real;
  Set	       = W.Set;
  String       = W.String;

  TEvent      	   = PROCEDURE {DELEGATE};
  TMouseEvent      = PROCEDURE {DELEGATE} (Sender: TObject; Shift: Set; Button,X,Y:Integer);
  TMouseMoveEvent  = PROCEDURE {DELEGATE} (Sender: TObject; Shift: Set; X,Y: Integer);
  TNotifyEvent     = PROCEDURE {DELEGATE} (Sender: TObject);
  TMouseWheelEvent = PROCEDURE {DELEGATE} (Sender: TObject; Shift: Set; Delta,X,Y: Integer);
  TKeyEvent 	   = PROCEDURE {DELEGATE} (Sender: TObject; VAR Key: ShortInt; Shift: Set);
  TKeyPressEvent   = PROCEDURE {DELEGATE} (Sender: TObject; VAR Key: Char);

  TObject* = POINTER TO RECORD END;

  TComponent*  = POINTER TO RECORD (TObject)
    handle*    : Integer;
    name*      : String;
    classname* : String;
    created*   : Boolean;
  END;

  TState*    = POINTER TO RECORD
    name*      : String;
    left*      : Integer;
    top*       : Integer;
    width*     : Integer;
    height*    : Integer;
    style*     : Set;
    exstyle*   : Set;
    caption*   : String;
    color*     : Integer;
    visible*   : Boolean;
    enabled*   : Boolean;
    align*     : Integer;
  END;

  TControl*    = POINTER TO RECORD (TComponent)
    id         : Integer;
    list*      : TList;
    descript*  : Set;
    left*      : Integer;
    top*       : Integer;
    width*     : Integer;
    height*    : Integer;
    parent*    : TControl;
    style*     : Set;
    exstyle*   : Set;
    caption*   : String;
    color*     : Integer;
    phndl*     : Integer;
    visible*   : Boolean;
    enabled*   : Boolean;
    state*     : TState;

    align*     : Integer;
    caTop      : TControl;
    caLeft     : TControl;
    caRight    : TControl;
    caBottom   : TControl;
    caClient   : TControl;

    (*property*)
    click	 : Boolean;
    showing	 : Boolean;
    OnMouseDown* : TMouseEvent;
    OnMouseUp*   : TMouseEvent;
    OnMouseMove* : TMouseMoveEvent;
    OnClick*     : TNotifyEvent;
    OnDblClick*  : TNotifyEvent;
    OnMouseWheel*: TMouseWheelEvent;
    OnResize*    : TEvent;
    OnPaint*	 : TEvent;
    OnKeyUp*	 : TKeyEvent;
    OnKeyDown*	 : TKeyEvent;
    OnKeyPress*	 : TKeyPressEvent;

    PROCEDURE & Create*(Parent: TControl);
    BEGIN
      NEW(list);
      id        := -1-ControlList.count;
      classname := "TControl";
      name      := "TControl";
      left      := 0;
      top       := 0;
      width     := 150;
      height    := 150;
      parent    := Parent;
      color	:= 0D9ECF0H;
      descript	:= CD_CONTROL;
      align     := 0;
      IF Parent = NIL THEN
        phndl := 0;
        IF MainForm = NIL THEN MainForm := SELF; END;
      ELSE
        phndl  := parent.handle;
      END;
      NEW(state);
      visible  := TRUE;
      enabled  := TRUE;
      showing  := TRUE;
      created  := FALSE;
      IF parent # NIL THEN
        handle := id;
        ControlList.Add(SELF);
      END;
    END Create;

    PROCEDURE CreateParams*;
    BEGIN
      style   := W.WS_CLIPCHILDREN + W.WS_CHILD +
		 W.WS_CLIPSIBLINGS + W.WS_BORDER;
      exstyle := W.WS_EX_TOOLWINDOW;
    END CreateParams;

    PROCEDURE CreateHandle*;
    BEGIN
      handle := W.CreateWindowEx(
         exstyle,classname,
         caption, style + W.WS_POPUP,
         left, top, width, height,
         0, 0, W.hInstance, 0
      );
    END CreateHandle;

    PROCEDURE SetHidePos;
    BEGIN
      W.SetWindowPos(
        handle, W.HWND_NOTOPMOST,
        left,top,width,height,
        W.SWP_HIDEWINDOW
      );
    END SetHidePos;

    PROCEDURE SetBounds;
    VAR
      R : W.RECT;
    BEGIN
      IF parent = NIL THEN
        W.GetWindowRect(handle,R);
        left   := R.left;
        top    := R.top;
        W.AdjustWindowRectEx(R,{},0,exstyle);
      ELSE
        W.GetClientRect(handle,R);
        W.AdjustWindowRectEx(R,style,0,exstyle);
      END;
      width  := R.right-R.left;
      height := R.bottom-R.top;
    END SetBounds;

    PROCEDURE SetShowPos;
    BEGIN
      W.SetWindowPos(
        handle, W.HWND_NOTOPMOST,
        left,top,width,height,
        W.SWP_SHOWWINDOW
      );
    END SetShowPos;

    PROCEDURE Update;
    BEGIN
      IF ~created THEN
        RegisterClass(classname);
        CreateParams;
        CreateHandle;
        SetBounds;
        SetHidePos;
        ControlList.Add(SELF);
        IF parent # NIL THEN
          parent.list.Add(SELF);
        END;
        W.UpdateWindow(handle);
        created  := TRUE;
      END;
    END Update;

    PROCEDURE Show*;
    BEGIN
      IF ~created THEN
        Update;
        IF parent = NIL THEN ControlList.CreateControl(SELF); END;
      END;
      W.ShowWindow(handle, W.SW_SHOW);
      W.UpdateWindow(handle);
      showing := TRUE;
    END Show;

    PROCEDURE Hide*;
    BEGIN
      W.ShowWindow(handle, W.SW_HIDE);
      W.UpdateWindow(handle);
      showing := FALSE;
    END Hide;

    PROCEDURE RegisterClass(ClassName: String);
    VAR
      class: W.WNDCLASSEX;
    BEGIN
      IF Cl.ClassList.Find(ClassName) THEN RETURN; END;
      class.cbSize        := SIZE(W.WNDCLASSEX);
      class.style         := W.CS_VREDRAW + W.CS_HREDRAW + W.CS_DBLCLKS;
      class.lpfnWndProc   := WndProc;
      class.cbClsExtra    := 0;
      class.cbWndExtra    := 0;
      class.hInstance     := W.hInstance;
      class.hIcon         := W.LoadIcon(0,SYSTEM.VAL(W.LPSTR,1));
      class.hCursor       := W.LoadCursor(0,SYSTEM.VAL(W.LPSTR,W.IDC_ARROW));
      class.hbrBackground := W.CreateSolidBrush(color);
      class.lpszMenuName  := W.NULL;
      class.lpszClassName := SYSTEM.ADR(ClassName);
      class.hIconSm       := W.LoadIcon(W.hInstance,SYSTEM.VAL(W.LPSTR,1));
      W.RegisterClassEx(class);
      Cl.ClassList.Add(ClassName,FALSE);
    END RegisterClass;

    (*-----------------------------------------------*)

    PROCEDURE MouseDown*(Shift: Set; Button,X,Y: Integer);
    BEGIN
      IF OnMouseDown # NIL THEN OnMouseDown(SELF,Shift,Button,X,Y); END;
    END MouseDown;

    PROCEDURE MouseUp*(Shift: Set; Button,X,Y: Integer);
    BEGIN
      IF OnMouseUp # NIL THEN OnMouseUp(SELF,Shift,Button,X,Y); END;
    END MouseUp;

    PROCEDURE MouseMove*(Shift: Set; X,Y: Integer);
    BEGIN
      IF OnMouseMove # NIL THEN OnMouseMove(SELF,Shift,X,Y); END;
    END MouseMove;

    PROCEDURE Click*;
    BEGIN
      IF OnClick # NIL THEN OnClick(SELF); END;
    END Click;

    PROCEDURE DblClick*;
    BEGIN
      IF OnDblClick # NIL THEN OnDblClick(SELF); END;
    END DblClick;

    PROCEDURE MouseWheel*(Shift: Set; Delta: Integer; X,Y: Integer);
    BEGIN
      IF OnMouseWheel # NIL THEN
        OnMouseWheel(SELF,Shift,Delta,X,Y);
      END;
    END MouseWheel;

    PROCEDURE KeyUp*(key: ShortInt; shift: Set);
    BEGIN
      IF OnKeyUp # NIL THEN OnKeyUp(SELF,key,shift); END;
    END KeyUp;

    PROCEDURE KeyDown*(key: ShortInt; shift: Set);
    BEGIN
      IF OnKeyDown # NIL THEN OnKeyDown(SELF,key,shift); END;
    END KeyDown;

    PROCEDURE KeyPress*(key: Char);
    BEGIN
      IF OnKeyPress # NIL THEN OnKeyPress(SELF,key); END;
    END KeyPress;

    PROCEDURE Resize*;
    BEGIN
      IF OnResize # NIL THEN OnResize; END;
    END Resize;

    PROCEDURE Paint*;
    BEGIN
      IF OnPaint # NIL THEN OnPaint; END;
    END Paint;

    PROCEDURE WMMouseDown*(button, wParam, lParam: Integer);
    VAR
      X,Y: Integer;
      key: Set;
    BEGIN
      X := LoWord(lParam);
      Y := HiWord(lParam);
      key := SysKeyToSet(wParam);
      click:= TRUE;
      MouseDown(key,button,X,Y);
    END WMMouseDown;

    PROCEDURE WMMouseUp*(button, wParam, lParam: Integer);
    VAR
      X,Y: Integer;
      key: Set;
    BEGIN
      X := LoWord(lParam);
      Y := HiWord(lParam);
      key := SysKeyToSet(wParam);
      MouseUp(key,button,X,Y);
      IF click THEN
        Click;
      END;
      click:= FALSE;
    END WMMouseUp;

    PROCEDURE WMMouseMove*(wParam, lParam: Integer);
    VAR
      X,Y: Integer;
      key: Set;
    BEGIN
      X := LoWord(lParam);
      Y := HiWord(lParam);
      key := SysKeyToSet(wParam);
      MouseMove(key,X,Y);
    END WMMouseMove;

    PROCEDURE WMDblClick*(button, wParam, lParam: Integer);
    BEGIN
      DblClick;
    END WMDblClick;

    PROCEDURE WMMouseWheel*(wParam, lParam: Integer);
    VAR
      X,Y, delta: Integer;
      key: Set;
    BEGIN
      X := LoWord(lParam);
      Y := HiWord(lParam);
      delta := HiWord(wParam) DIV 120;
      key := SysKeyToSet(wParam);
      MouseWheel(key,delta,X,Y);
    END WMMouseWheel;

    PROCEDURE WMKeyDown*(wParam, lParam: Integer);
    BEGIN
      KeyDown(SHORT(wParam),SYSTEM.VAL(Set,lParam));
    END WMKeyDown;

    PROCEDURE WMKeyUp*(wParam, lParam: Integer);
    BEGIN
      KeyUp(SHORT(wParam),SYSTEM.VAL(Set,lParam));
    END WMKeyUp;

    PROCEDURE WMChar*(wParam, lParam: Integer);
    BEGIN
      KeyPress(CHR(SHORT(LoWord(wParam))));
    END WMChar;

    PROCEDURE WMPaint*(wParam, lParam: Integer);
    VAR
      dc: Integer;
      ps: W.PAINTSTRUCT;
      rect: W.RECT;
      left,top,width,height: Integer;
      brush : W.HBRUSH;
    BEGIN
      W.GetClientRect(handle,rect);
      left := rect.left;
      top := rect.top;
      width := rect.right-rect.left;
      height := rect.bottom-rect.top;
      dc:= W.BeginPaint(handle,ps);
      brush  := W.CreateSolidBrush(color);
      W.SelectObject(dc, brush);
      W.PatBlt(dc,left,top,width,height,W.PATCOPY);
      W.DeleteObject(brush);
      Paint;
      W.EndPaint(handle,ps);
      W.ReleaseDC(dc,handle);
    END WMPaint;

    PROCEDURE WMSize*(wParam, lParam: Integer);
    VAR
      rect: W.RECT;
    BEGIN
      SetBounds;
      Resize;

      W.GetClientRect(handle,rect);

      IF caTop # NIL THEN
        caTop.left := rect.left;
        caTop.width := rect.right-rect.left;
        caTop.top := rect.top;
      END;

      IF caBottom # NIL THEN
        caBottom.left := rect.left;
        caBottom.width := rect.right-rect.left;
        caBottom.top := rect.bottom - rect.top - caBottom.height;
      END;

      IF caLeft # NIL THEN
        caLeft.left := rect.left;
        IF caTop # NIL THEN
          caLeft.top  := caTop.height;
        ELSE
          caLeft.top  := rect.top;
        END;
        IF caBottom # NIL THEN
          caLeft.height  := caBottom.top - caLeft.top;
        ELSE
          caLeft.height  := rect.bottom - rect.top;
        END;
      END;

      IF caRight # NIL THEN
        caRight.left := rect.right-rect.left-caRight.width;
        IF caTop # NIL THEN
          caRight.top  := caTop.height;
        ELSE
          caRight.top  := rect.top;
        END;
        IF caBottom # NIL THEN
          caRight.height  := caBottom.top - caRight.top;
        ELSE
          caRight.height  := rect.bottom - rect.top;
        END;
      END;

      IF caClient # NIL THEN
        IF caTop # NIL THEN
          caClient.top  := caTop.height;
        ELSE
          caClient.top  := rect.top;
        END;

        IF caLeft # NIL THEN
          caClient.left  := caLeft.width;
        ELSE
          caClient.left  := rect.left;
        END;

        IF caRight # NIL THEN
          caClient.width  := caRight.left-caClient.left;
        ELSE
          caClient.width  := rect.right-rect.left-caClient.left;
        END;

        IF caBottom # NIL THEN
          caClient.height  := caBottom.top - caClient.top;
        ELSE
          caClient.height  := rect.bottom - caClient.top;
        END;
      END;

      list.CheckState(NIL);
    END WMSize;

    PROCEDURE WMMove*(wParam, lParam: Integer);
    BEGIN
      SetBounds;
      Resize;
      list.CheckState(NIL);
    END WMMove;

    (*-----------------------------------------------*)

    PROCEDURE ControlProc*(VAR message, wParam, lParam: Integer): Integer;
    BEGIN

      IF    message = W.WM_LBUTTONDOWN   THEN WMMouseDown(mbLeft,wParam,lParam)
      ELSIF message = W.WM_RBUTTONDOWN   THEN WMMouseDown(mbRight,wParam,lParam)
      ELSIF message = W.WM_MBUTTONDOWN   THEN WMMouseDown(mbMiddle,wParam,lParam)
      ELSIF message = W.WM_LBUTTONUP     THEN WMMouseUp(mbLeft,wParam,lParam)
      ELSIF message = W.WM_RBUTTONUP     THEN WMMouseUp(mbRight,wParam,lParam)
      ELSIF message = W.WM_MBUTTONUP     THEN WMMouseUp(mbMiddle,wParam,lParam)
      ELSIF message = W.WM_MOUSEMOVE     THEN WMMouseMove(wParam,lParam)
      ELSIF message = W.WM_LBUTTONDBLCLK THEN WMDblClick(mbLeft,wParam,lParam)
      ELSIF message = W.WM_RBUTTONDBLCLK THEN WMDblClick(mbRight,wParam,lParam)
      ELSIF message = W.WM_MBUTTONDBLCLK THEN WMDblClick(mbMiddle,wParam,lParam)
      ELSIF message = W.WM_MOUSEWHEEL    THEN WMMouseWheel(wParam,lParam)
      ELSIF message = W.WM_KEYDOWN       THEN WMKeyDown(wParam,lParam)
      ELSIF message = W.WM_KEYUP         THEN WMKeyUp(wParam,lParam)
      ELSIF message = W.WM_CHAR	         THEN WMChar(wParam,lParam)
      ELSIF message = W.WM_SIZE	         THEN WMSize(wParam,lParam)
      ELSIF message = W.WM_MOVE	         THEN WMMove(wParam,lParam)
      ELSIF message = W.WM_PAINT         THEN WMPaint(wParam,lParam)
      ELSIF message = W.WM_COMMAND       THEN
      ELSIF message = W.WM_ACTIVATE      THEN
      ELSIF message = W.WM_ERASEBKGND    THEN RETURN 1
      ELSIF message = W.WM_CLOSE 	 THEN
                                              IF descript = CD_WINDOW THEN
                                                IF SELF = MainForm THEN
                                                  W.PostQuitMessage(0);
                                                ELSE
                                                  SELF.Hide;
                                                END;
                                              END;
      ELSE
        RETURN -1
      END;
      RETURN 0;
    END ControlProc;

  END;

  (*########################################*)
  (******************************************)
  (*########################################*)

  TList*  = POINTER TO RECORD
    left     : TList;
    right    : TList;
    control* : TControl;
    idx      : Integer;
    count    : Integer;

    PROCEDURE & Init*;
    BEGIN
      left    := NIL;
      right   := NIL;
      idx     := -1;
      control := NIL;
      count   := 0;
    END Init;

    PROCEDURE Add*(control: TControl);
    VAR
      ob0, ob1: TList;
      left: Boolean;
    BEGIN

      ob0 := SELF;
      ob1 := ob0.right;
      left := FALSE;

      LOOP
        IF ob1 # NIL THEN
          IF control.handle < ob1.idx THEN
	    ob0 := ob1;
	    ob1 := ob0.left;
	    left := TRUE
          ELSIF control.handle > ob1.idx THEN
	    ob0 := ob1;
	    ob1 := ob0.right;
	    left := FALSE
          ELSE
            EXIT
          END;
        ELSE (*insert*)
          NEW(ob1);
          IF left THEN
	    ob0.left := ob1
          ELSE
	    ob0.right := ob1
          END;
          ob1.left:=NIL;
          ob1.right:=NIL;
          ob1.control := control;
          ob1.idx := control.handle;
          INC(count);
          EXIT;
        END;
      END;

    END Add;

    PROCEDURE Find*(idx: Integer): TControl;
    VAR
      obj: TList;
    BEGIN
      obj := SELF;
      LOOP
        IF obj # NIL THEN
          IF idx < obj.idx THEN
	    obj := obj.left;
          ELSIF idx > obj.idx THEN
	    obj := obj.right;
          ELSE (*ok*)
            EXIT
          END;
        ELSE (*no found*)
          RETURN NIL;
        END;
      END;
      RETURN obj.control;
    END Find;

    PROCEDURE SaveState*(control: TControl);

      PROCEDURE Save(VAR obj: TList);
      VAR
        ctl: TControl;
      BEGIN
        IF control = NIL THEN
          IF obj = NIL THEN RETURN END;
          IF obj.left  # NIL THEN Save(obj.left); END;
          IF obj.right # NIL THEN Save(obj.right); END;
          ctl := obj.control;
        ELSE
          ctl := control;
        END;
        ctl.state.name      := ctl.name;
        ctl.state.left      := ctl.left;
        ctl.state.top       := ctl.top;
        ctl.state.width     := ctl.width ;
        ctl.state.height    := ctl.height;
        ctl.state.style     := ctl.style ;
        ctl.state.exstyle   := ctl.exstyle;
        ctl.state.caption   := ctl.caption;
        ctl.state.color     := ctl.color;
        ctl.state.visible   := ctl.visible;
        ctl.state.enabled   := ctl.enabled;
        ctl.state.align     := ctl.align;
      END Save;

    BEGIN
      IF control = NIL THEN
        Save(left);
        Save(right);
      ELSE
        Save(NIL);
      END;
    END SaveState;

    PROCEDURE CheckState*(control: TControl);
    VAR
      data: Set;
      offset: Integer;

      PROCEDURE Check(VAR obj: TList);
      VAR
        ctl: TControl;
      BEGIN
        IF control = NIL THEN
          IF obj = NIL THEN RETURN END;
          IF obj.left  # NIL THEN Check(obj.left); END;
          IF obj.right # NIL THEN Check(obj.right); END;
          ctl := obj.control;
        ELSE
          ctl := control;
        END;

        data := {};

        IF ctl.state.name      # ctl.name        THEN INCL(data,1) END;
        IF ctl.state.left      # ctl.left        THEN INCL(data,4) END;
        IF ctl.state.top       # ctl.top         THEN INCL(data,4) END;
        IF ctl.state.width     # ctl.width       THEN INCL(data,4) END;
        IF ctl.state.height    # ctl.height      THEN INCL(data,4) END;
        IF ctl.state.style     # ctl.style       THEN INCL(data,10) END;
        IF ctl.state.exstyle   # ctl.exstyle     THEN INCL(data,11) END;
        IF ctl.state.caption   # ctl.caption     THEN INCL(data,12) END;
        IF ctl.state.color     # ctl.color       THEN INCL(data,13) END;
        IF ctl.state.visible   # ctl.visible     THEN INCL(data,14) END;
        IF ctl.state.enabled   # ctl.enabled     THEN INCL(data,15) END;
        IF ctl.state.align     # ctl.align       THEN INCL(data,17) END;

        IF (data # {}) THEN

          IF 4 IN data THEN
            W.MoveWindow(
	      ctl.handle,
	      ctl.left,ctl.top,
	      ctl.width,ctl.height,
	      1);
          END;

          IF 12 IN data THEN
            W.SetWindowText(ctl.handle,SYSTEM.ADR(ctl.caption));
          END;

          IF 13 IN data THEN
            W.InvalidateRect( ctl.handle,NIL,FALSE);
          END;

          IF 14 IN data THEN
            IF ~ ctl.visible THEN
              IF ctl.showing THEN ctl.Hide; END;
            ELSE
              IF ~ ctl.showing THEN ctl.Show; END;
            END;
          END;

          IF 15 IN data THEN
            IF ctl.enabled THEN
              W.EnableWindow(ctl.handle,1);
            ELSE
              W.EnableWindow(ctl.handle,0);
            END;
          END;

          IF 17 IN data THEN
            IF ctl.parent # NIL THEN
              IF alTop    = ctl.align THEN ctl.parent.caTop    := ctl END;
              IF alLeft   = ctl.align THEN ctl.parent.caLeft   := ctl END;
              IF alRight  = ctl.align THEN ctl.parent.caRight  := ctl END;
              IF alBottom = ctl.align THEN ctl.parent.caBottom := ctl END;
              IF alClient = ctl.align THEN ctl.parent.caClient := ctl END;
            END;
          END;

        END;

      END Check;

    BEGIN
      IF control = NIL THEN
        Check(left);
        Check(right);
      ELSE
        Check(NIL);
      END;
    END CheckState;

    PROCEDURE Update*(control: TControl);
    BEGIN
      CheckState(control);
      SaveState(control);
    END Update;

    PROCEDURE CreateControl*(control: TControl);

      PROCEDURE Create(VAR obj: TList);
      VAR
        st: Set;
      BEGIN
          IF obj = NIL THEN RETURN END;
          IF ~obj.control.created THEN
            obj.control.Update;
            st := W.GetWindowLong(obj.control.parent.handle,W.GWL_STYLE);
            W.SetWindowLong(obj.control.parent.handle,W.GWL_STYLE,st+W.WS_POPUP);
            W.SetParent(obj.control.handle,obj.control.parent.handle);
            IF obj.control.parent.parent # NIL THEN
              W.SetWindowLong(obj.control.parent.handle,W.GWL_STYLE,st-W.WS_POPUP+W.WS_CHILD);
              W.SetWindowLong(obj.control.handle,W.GWL_STYLE,st-W.WS_POPUP+W.WS_CHILD);
            END;
            st := W.GetWindowLong(obj.control.handle,W.GWL_STYLE);
            W.SetWindowLong(obj.control.handle,W.GWL_STYLE,st-W.WS_POPUP+W.WS_CHILD);
            IF obj.control.visible THEN obj.control.Show END;
            ControlList.Update(obj.control);
          END;
          IF obj.left  # NIL THEN Create(obj.left); END;
          IF obj.right # NIL THEN Create(obj.right); END;
      END Create;

    BEGIN
      Create(left);
      Create(right);
      control.width:=control.width+1;
      ControlList.Update(control);
      control.width:=control.width-1;
    END CreateControl;

  END;

PROCEDURE [WINAPI] WndProc* (window: W.HWND; message: Integer; wParam: W.WPARAM; lParam: W.LPARAM): W.LRESULT;
VAR
  cl: TControl;
  res: Integer;
BEGIN
  res := -1;
  cl := ControlList.Find(window);
  IF cl # NIL THEN
    res := cl.ControlProc(message, wParam, lParam);
  END;
  IF res = -1 THEN
    RETURN W.DefWindowProc(window, message, wParam, lParam)
  END;
  RETURN res
END WndProc;

(*###############################################*)

PROCEDURE LoWord(VAR n: Integer): Integer;
BEGIN
  RETURN ASH(SYSTEM.LSH(n, 16), -16)
END LoWord;

PROCEDURE HiWord(VAR n: Integer): Integer;
BEGIN
  RETURN ASH(n, -16);
END HiWord;

PROCEDURE SysKeyToSet(Param: Integer): Set;
VAR
  key, sysKey: Set;
BEGIN
  key := {};
  sysKey := SYSTEM.VAL(Set,Param);
  IF 2 IN sysKey THEN INCL(key,ssShift); END;
  IF 3 IN sysKey THEN INCL(key,ssCtrl); END;
  RETURN key;
END SysKeyToSet;

PROCEDURE Run*;
VAR
  msg : W.MSG;
BEGIN
  WHILE W.GetMessage(msg, 0, 0, 0) # 0 DO
    W.TranslateMessage(msg);
    W.DispatchMessage(msg);
    ControlList.Update(NIL);
  END;
  Cl.ClassList.UnregisterAllClass;
  W.Shutdown();
END Run;

(*#######################################*)
(*#######################################*)
(*#######################################*)

BEGIN
  NEW(ControlList);
  MainForm := NIL;
END Controls.

