MODULE Files;

IMPORT
  SYSTEM, W := Windows;

TYPE
  Byte         = W.Byte;
  Char         = W.Char;
  Boolean      = W.Boolean;
  SmallInt     = W.SmallInt;
  ShortInt     = W.ShortInt;
  Integer      = W.Integer;
  Longint      = W.Longint;
  Shortreal    = W.Shortreal;
  Real         = W.Real;
  Set	       = W.Set;
  String       = W.String;

  Bytes4       = ARRAY 4 OF Byte;
  Bytes8       = ARRAY 8 OF Byte;

  TFile* = POINTER TO RECORD
    handle : Integer;
    name*  : String;
    path*  : String;
    len*   : Integer;
    pos*   : Integer;
    eof*   : Boolean;

    PROCEDURE & Init*(filename: String);
    BEGIN
      name := filename;
      path := filename;
      pos  := 0;
      len  := 0;
      eof  := FALSE;
    END Init;

    PROCEDURE Exists*(): Boolean;
    VAR
      res: Set;
    BEGIN
      res := W.GetFileAttributes(path);
      RETURN res # W.INVALID_FILE_ATTRIBUTES;
    END Exists;

    PROCEDURE Open*;
    VAR
      res: Integer;
    BEGIN
      IF Exists() THEN
        handle := W.CreateFile(
	            SYSTEM.ADR(path),
		    W.GENERIC_READ + W.GENERIC_WRITE
                    ,W.FILE_SHARE_READ+W.FILE_SHARE_WRITE+W.FILE_SHARE_DELETE, NIL,
		    W.OPEN_EXISTING, W.FILE_ATTRIBUTE_NORMAL, 0);
        SetPos(0);
        len := W.GetFileSize(handle,res);
      ELSE
        W.MessageBox(0,'File not existed!',"Warning",{});
      END
    END Open;

    PROCEDURE Create*;
    VAR
      res: Integer;
    BEGIN
      handle := W.CreateFile(
                  SYSTEM.ADR(path),
                  W.GENERIC_READ + W.GENERIC_WRITE,
                  {}, NIL,
                  W.CREATE_NEW,W.FILE_ATTRIBUTE_NORMAL, 0);
      pos := 0;
      len := 0;
    END Create;

    PROCEDURE Close*;
    BEGIN
      W.CloseHandle(handle);
    END Close;

    PROCEDURE SetPos*(position : Integer);
    VAR
      res : Integer;
    BEGIN
      W.SetFilePointer(handle,pos,res,W.FILE_BEGIN);
      pos := position;
    END SetPos;

    PROCEDURE ReadByte*(VAR x: Byte);
    VAR
      res,count : Integer;
      b: ARRAY 1 OF Byte;
    BEGIN
      IF pos < len THEN
        W.ReadFile(handle,b,1,count,NIL);
        INC(pos);
        x := b[0];
      ELSE
        x := 0;
        eof := TRUE
      END
    END ReadByte;

    PROCEDURE ReadBytes*(VAR x: ARRAY OF Byte; len: Integer);
    VAR
      b: Byte;
      i: Integer;
    BEGIN
      i:=0;
      WHILE ~ eof & (i<len) DO
        ReadByte(b);
        x[i]:=b;
        INC(i);
      END;
    END ReadBytes;

    PROCEDURE Read*(VAR x: Byte);
    BEGIN
      ReadByte(x);
    END Read;

    PROCEDURE ReadInt*(VAR x: ShortInt);
    VAR
      x0, x1: SmallInt;
    BEGIN
      Read(x0);
      Read(x1);
      x := LONG(x1) * 100H + LONG(x0) MOD 100H
    END ReadInt;

    PROCEDURE ReadLInt*(VAR x: Integer);
    BEGIN
      ReadBytes(SYSTEM.VAL(Bytes4, x), 4)
    END ReadLInt;

    PROCEDURE ReadSet*(VAR x: Set);
    BEGIN
      ReadBytes(SYSTEM.VAL(Bytes4, x), 4)
    END ReadSet;

    PROCEDURE ReadBool*(VAR x: Boolean);
    VAR
      s: SmallInt;
    BEGIN
      Read(s);
      x := s # 0
    END ReadBool;

    PROCEDURE ReadReal*(VAR x: Shortreal);
    BEGIN
      ReadBytes(SYSTEM.VAL(Bytes4, x), 4)
    END ReadReal;

    PROCEDURE ReadLReal*(VAR x: Real);
    BEGIN
      ReadBytes(SYSTEM.VAL(Bytes8, x), 8)
    END ReadLReal;

  END;

END Files.




