MODULE Images;

IMPORT
  SYSTEM, W := Windows;

TYPE
  Byte         = W.Byte;
  Char         = W.Char;
  Boolean      = W.Boolean;
  SmallInt     = W.SmallInt;
  ShortInt     = W.ShortInt;
  Integer      = W.Integer;
  Longint      = W.Longint;
  Shortreal    = W.Shortreal;
  Real         = W.Real;
  Set	       = W.Set;
  String       = W.String;

  TGraphics* = POINTER TO RECORD
    magic*   : Integer;
    handle*  : Integer;
    pointer* : Integer;
    width*   : Integer;
    height*  : Integer;
    bitsize* : Integer;
  END;

  TBitmap* = POINTER TO RECORD (TGraphics)
    info    : W.BITMAP;
    memdc   : Integer;
    BI      : W.BITMAPINFO;

    PROCEDURE & Init;
    BEGIN
      BI.bmiHeader.biSize        := SIZE(W.BITMAPINFOHEADER);
      BI.bmiHeader.biWidth       := 0;
      BI.bmiHeader.biHeight      := 0;
      BI.bmiHeader.biPlanes      := 1;
      BI.bmiHeader.biBitCount    := 32;
      BI.bmiHeader.biCompression := W.BI_RGB;
      memdc := W.CreateCompatibleDC(0);
      bitsize := BI.bmiHeader.biBitCount DIV 8;
      width  := 0;
      height := 0;
      handle := 0;
    END Init;

    PROCEDURE Create*(w,h: Integer);
    BEGIN
      BI.bmiHeader.biWidth       := w;
      BI.bmiHeader.biHeight      := -h;
      width  := w;
      height := h;
      handle := W.CreateDIBSection(0, BI, W.DIB_RGB_COLORS, pointer, 0, 0);
      SetMagic;
    END Create;

    PROCEDURE SetMagic;
    BEGIN
      SYSTEM.GET(pointer,magic);
    END SetMagic;

    PROCEDURE LoadFromFile*(name: String);
    VAR
      mem: Integer;
      bmp: Integer;
    BEGIN
      bmp := W.LoadImage(0, SYSTEM.ADR(name), W.IMAGE_BITMAP, 0, 0, W.LR_LOADFROMFILE);
      W.GetObject(bmp,SIZE(W.BITMAP),SYSTEM.ADR(info));
      BI.bmiHeader.biWidth       := info.bmWidth;
      BI.bmiHeader.biHeight      := -info.bmHeight;
      width                      := info.bmWidth;
      height                     := info.bmHeight;
      mem                        := W.CreateCompatibleDC(0);
      W.SelectObject(mem,bmp);
      handle := W.CreateDIBSection(0, BI, W.DIB_RGB_COLORS, pointer, 0, 0);
      W.GetDIBits(mem, bmp, 0, info.bmHeight, pointer, BI, W.DIB_RGB_COLORS);
      SetMagic;
      W.DeleteObject(bmp);
      W.DeleteDC(mem);
    END LoadFromFile;

    PROCEDURE SetPixel*(x,y, c: Integer);
    BEGIN
      IF (x>=0) & (x<width) & (y>=0) & (y<height) THEN
        SYSTEM.PUT(pointer+(x+y*width)*bitsize,c);
      END;
    END SetPixel;

    PROCEDURE Free*;
    BEGIN
      W.DeleteObject(handle);
      W.DeleteDC(memdc);
    END Free;

  END;

  TImageList* = POINTER TO RECORD
    bitmap  : TBitmap;
    width*  : Integer;
    height* : Integer;
    item    : TGraphics;
    pixels  : POINTER TO ARRAY OF Integer;
    count*  : Integer;

    PROCEDURE & Init (w,h: Integer);
    BEGIN
      NEW(bitmap);
      width  := w;
      height := h;
      NEW(pixels,w*h);
      count := 0;
    END Init;

    PROCEDURE LoadFromFile*(name: String);
    BEGIN
      bitmap.LoadFromFile(name);
      count := ((bitmap.height) DIV height)*((bitmap.width) DIV width);
    END LoadFromFile;

    PROCEDURE GetImage*(idx: Integer): TGraphics;
    VAR
      x,y,j: Integer;
      ss: String;
    BEGIN
      IF (count = 0) OR (idx>=count) OR (height*width=0) THEN RETURN NIL END;

      y:=idx DIV ((bitmap.width+width-1) DIV width);
      x:=idx-y*((bitmap.width+width-1) DIV width);
      y:=y*height;
      x:=x*width;

      NEW(item);
      item.width   := width;
      item.height  := height;
      item.bitsize := bitmap.bitsize;
      item.magic   := bitmap.magic;
      FOR j:=0 TO height-1 DO
        SYSTEM.MOVE(
              bitmap.pointer+(x+y*bitmap.width)*bitmap.bitsize
              +(j*bitmap.width)*bitmap.bitsize,
              SYSTEM.ADR(pixels[j*width]),
              width*bitmap.bitsize);
      END;
      item.pointer := SYSTEM.VAL(Integer,pixels)+4*bitmap.bitsize;
      RETURN item;
    END GetImage;

    PROCEDURE Free*;
    BEGIN
      bitmap.Free;
      bitmap := NIL;
      pixels := NIL;
    END Free;

  END;

END Images.
